"use strict";
/**
 * @license
 * Copyright Google Inc. All Rights Reserved.
 *
 * Use of this source code is governed by an MIT-style license that can be
 * found in the LICENSE file at https://angular.io/license
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.pathHasSupportedExtension = exports.generateTsMigrationExportsShimFile = void 0;
/**
 * Processes goog.tsMigrationExportsShim statements into
 * ".legacy.closure.js" and ".legacy.d.ts" files.
 */
const ts = require("typescript");
const transformer_util_1 = require("./transformer_util");
/** Silence linter. */
function generateTsMigrationExportsShimFile(src, typeChecker, host, manifest) {
    const srcFilename = host.rootDirsRelative(src.fileName);
    const srcModuleId = host.pathToModuleName('', src.fileName);
    const srcIds = new FileIdGroup(srcFilename, srcModuleId);
    return new Generator(src, typeChecker, host, manifest, srcIds).run();
}
exports.generateTsMigrationExportsShimFile = generateTsMigrationExportsShimFile;
/**
 * Does this path have a file extension that supports
 * goog.tsMigrationExportsShim?
 */
function pathHasSupportedExtension(p) {
    return Boolean(p.match(SUPPORTED_EXTENSIONS));
}
exports.pathHasSupportedExtension = pathHasSupportedExtension;
function stripSupportedExtensions(path) {
    return path.replace(SUPPORTED_EXTENSIONS, '');
}
// .ts but not .d.ts
const SUPPORTED_EXTENSIONS = /(?<!\.d)\.ts$/;
/**
 * A one-time-use object for generating tsMigrationExportsShim output files.
 *
 * The generator is able to emit two different outputs: .js and .d.ts. Only one
 * of thse should be produced for a given src file. The choice of which depends
 * on whether tsickle is producing code for downstream typechecking (.d.ts file
 * for TS libraries) or optimization (.js files for Closure Compiler).
 */
class Generator {
    constructor(src, typeChecker, host, manifest, srcIds) {
        this.src = src;
        this.typeChecker = typeChecker;
        this.host = host;
        this.manifest = manifest;
        this.srcIds = srcIds;
        this.diagnostics = [];
        const moduleSymbol = this.typeChecker.getSymbolAtLocation(this.src);
        this.mainExports =
            moduleSymbol ? this.typeChecker.getExportsOfModule(moduleSymbol) : [];
    }
    run() {
        const outputFilename = this.srcIds.google3PathWithoutExtension() +
            (this.host.transformTypesToClosure ? '.tsmes.closure.js' :
                '.tsmes.d.ts');
        const tsmessBreakdown = this.extractTsmesStatement();
        if (!tsmessBreakdown) {
            return {
                filename: outputFilename,
                content: '',
                diagnostics: this.diagnostics,
            };
        }
        this.outputIds =
            new FileIdGroup(outputFilename, tsmessBreakdown.googModuleId);
        this.googExports = tsmessBreakdown.googExports;
        return {
            filename: outputFilename,
            content: this.host.transformTypesToClosure ? this.generateTsmesJs() :
                this.generateTsmesDts(),
            diagnostics: this.diagnostics,
        };
    }
    /**
     * Finds the top-level call to tsmes in the input, if any,
     * and returns the relevant info from within.
     *
     * If no such call exists, or the call is malformed, returns undefined.
     * Diagnostics about malformed calls will also be logged.
     */
    extractTsmesStatement() {
        let tsmesCall = undefined;
        for (const statement of this.src.statements) {
            if (!ts.isExpressionStatement(statement) ||
                !transformer_util_1.isTsMigrationExportsShimCall(statement.expression)) {
                this.checkNonTopLevelTsmesCalls(statement);
                continue;
            }
            const call = statement.expression;
            if (tsmesCall) {
                this.report(call, 'at most one goog.tsMigrationExportsShim is allowed per file');
            }
            else {
                tsmesCall = call;
            }
        }
        if (!tsmesCall) {
            return undefined;
        }
        if (tsmesCall.arguments.length !== 2) {
            this.report(tsmesCall, 'goog.tsMigrationExportsShim requires 2 arguments');
            return undefined;
        }
        const [moduleId, exportsExpr] = tsmesCall.arguments;
        if (!ts.isStringLiteral(moduleId)) {
            this.report(moduleId, 'goog.tsMigrationExportsShim ID must be a string literal');
            return undefined;
        }
        const googExports = this.extractGoogExports(exportsExpr);
        if (!googExports) {
            return undefined;
        }
        return {
            googModuleId: moduleId.text,
            googExports,
        };
    }
    /**
     * Given the exports from a tsmes call, return a simplified model of the
     * relevant values.
     *
     * If the exports are malformed, returns undefined. Diagnostics about
     * malformed exports are also logged.
     */
    extractGoogExports(exportsExpr) {
        let googExports;
        const diagnosticCount = this.diagnostics.length;
        if (ts.isObjectLiteralExpression(exportsExpr)) {
            googExports = new Map();
            for (const property of exportsExpr.properties) {
                if (ts.isShorthandPropertyAssignment(property)) {
                    // {Bar}
                    const symbol = this.typeChecker.getShorthandAssignmentValueSymbol(property);
                    this.checkIsModuleExport(property.name, symbol);
                    googExports.set(property.name.text, property.name.text);
                }
                else if (ts.isPropertyAssignment(property)) {
                    // {Foo: Bar}
                    const name = property.name;
                    if (!ts.isIdentifier(name)) {
                        this.report(name, 'export names must be simple keys');
                        continue;
                    }
                    const init = property.initializer;
                    if (!ts.isIdentifier(init)) {
                        this.report(init, 'export values must be plain identifiers');
                        continue;
                    }
                    const symbol = this.typeChecker.getSymbolAtLocation(init);
                    this.checkIsModuleExport(init, symbol);
                    googExports.set(name.text, init.text);
                }
                else {
                    this.report(property, `exports object must only contain (shorthand) properties`);
                }
            }
        }
        else if (ts.isIdentifier(exportsExpr)) {
            const symbol = this.typeChecker.getSymbolAtLocation(exportsExpr);
            this.checkIsModuleExport(exportsExpr, symbol);
            googExports = exportsExpr.text;
        }
        else {
            this.report(exportsExpr, `exports object must be either an object literal ({A, B}) or the ` +
                `identifier of a module export (A)`);
        }
        return (diagnosticCount === this.diagnostics.length) ? googExports :
            undefined;
    }
    /**
     * Recurse through top-level statments looking for tsmes calls.
     *
     * tsmes is only allowed as a top-level statement, so if we find it deeper
     * down we report an error.
     */
    checkNonTopLevelTsmesCalls(topLevelStatement) {
        const inner = (node) => {
            if (transformer_util_1.isTsMigrationExportsShimCall(node)) {
                this.report(node, 'goog.tsMigrationExportsShim is only allowed in top level statements');
            }
            ts.forEachChild(node, inner);
        };
        ts.forEachChild(topLevelStatement, inner);
    }
    /**
     * Generate the JS file that other JS files will goog.require to use the
     * shimmed export layout.
     */
    generateTsmesJs() {
        if (!this.outputIds || !this.googExports) {
            throw new Error();
        }
        const mainRequireImports = this.mainExports.map((e) => e.name).join(', ');
        const mainModuleRequire = `const { ${mainRequireImports} } = ` +
            `goog.require('${this.srcIds.googModuleId}');`;
        let exportsAssignment;
        if (this.googExports instanceof Map) {
            // In the case that tsmes was passed named exports.
            const bindings = Array.from(this.googExports).map(([k, v]) => `  ${k}: ${v},`);
            exportsAssignment = lines(`exports = {`, ...bindings, `};`);
        }
        else {
            // In the case that tsmes was passed a default export.
            exportsAssignment = `exports = ${this.googExports};`;
        }
        this.manifest.addModule(this.outputIds.google3Path, this.outputIds.googModuleId);
        this.manifest.addReferencedModule(this.outputIds.google3Path, this.srcIds.googModuleId);
        return lines('/**', ' * @fileoverview generator:ts_migration_exports_shim.ts', ' * generated_from:' + this.srcIds.google3Path, ' */', `goog.module('${this.outputIds.googModuleId}');`, mainModuleRequire, exportsAssignment, '');
    }
    /**
     * Generate the .d.ts file that approximates what clutz would generate for the
     * file produced by generateTsmesJs.
     *
     * Since no JS library holds the generated JS file, clutz will never run over
     * that code. This .d.ts is needed for downstream TS libraries to know about
     * the shimmed export types.
     */
    generateTsmesDts() {
        if (!this.outputIds || !this.googExports) {
            throw new Error();
        }
        const generatedFromComment = '// Generated from ' + this.srcIds.google3Path;
        const dependencyFileImports = lines(`declare module 'ಠ_ಠ.clutz._dependencies' {`, `  import '${this.srcIds.esModuleImportPath()}';`, `}`);
        let clutzNamespaceDeclaration;
        let googColonModuleDeclaration;
        if (this.googExports instanceof Map) {
            // In the case that tsmes was passed named exports.
            const clutzNamespace = this.srcIds.clutzNamespace();
            const clutzNamespaceReexports = Array.from(this.googExports)
                .map(([k, v]) => `  export import ${k} = ${clutzNamespace}.${v};`);
            clutzNamespaceDeclaration = lines(generatedFromComment, `declare namespace ${this.outputIds.clutzNamespace()} {`, ...clutzNamespaceReexports, `}`);
            googColonModuleDeclaration = lines(generatedFromComment, `declare module '${this.outputIds.clutzModuleId()}' {`, `  import x = ${this.outputIds.clutzNamespace()};`, `  export = x;`, `}`);
        }
        else {
            // In the case that tsmes was passed a default export.
            clutzNamespaceDeclaration = lines(generatedFromComment, `declare namespace ಠ_ಠ.clutz {`, `  export import ${this.outputIds.googModuleRewrittenId()} =`, `      ${this.srcIds.clutzNamespace()}.${this.googExports};`, `}`);
            googColonModuleDeclaration = lines(generatedFromComment, `declare module '${this.outputIds.clutzModuleId()}' {`, `  import x = ${this.outputIds.clutzNamespace()};`, `  export default x;`, `}`);
        }
        return lines('/**', ' * @fileoverview generator:ts_migration_exports_shim.ts', ' */', dependencyFileImports, clutzNamespaceDeclaration, googColonModuleDeclaration, '');
    }
    checkIsModuleExport(node, symbol) {
        if (!symbol) {
            this.report(node, `could not resolve symbol of exported property`);
        }
        else if (this.mainExports.indexOf(symbol) === -1) {
            this.report(node, `export must be an exported symbol of the module`);
        }
        else {
            return true;
        }
        return false;
    }
    report(node, messageText, textRange, category = ts.DiagnosticCategory.Error) {
        transformer_util_1.reportDiagnostic(this.diagnostics, node, messageText, textRange, category);
    }
}
function lines(...x) {
    return x.join('\n');
}
/**
 * The set of IDs associated with a single file.
 *
 * Each file can be identified in multiple ways, many of which are derivatives
 * of one another.
 */
class FileIdGroup {
    constructor(google3Path, googModuleId) {
        this.google3Path = google3Path;
        this.googModuleId = googModuleId;
    }
    google3PathWithoutExtension() {
        return stripSupportedExtensions(this.google3Path);
    }
    esModuleImportPath() {
        return 'google3/' + this.google3PathWithoutExtension();
    }
    googModuleRewrittenId() {
        return 'module$exports$' + this.googModuleId.replace(/\./g, '$');
    }
    clutzNamespace() {
        return 'ಠ_ಠ.clutz.' + this.googModuleRewrittenId();
    }
    clutzModuleId() {
        return 'goog:' + this.googModuleId;
    }
}
//# sourceMappingURL=ts_migration_exports_shim.js.map